int sys32_newstat(void) {return 0;}
int sys32_newlstat(void) {return 0;}
int sys32_newfstat(void) {return 0;}
int sys_mmap2(void) {return 0;}
int sys_truncate64(void) {return 0;}
int sys_ftruncate64(void) {return 0;}
int sys_stat64(void) {return 0;}
int sys_lstat64(void) {return 0;}
int sys_fstat64(void) {return 0;}

